# validated: 2019-01-04 DS ceed1d74dc30 cpp/Dispatcher.cpp cpp/Dispatcher.h cpp/IDispatcher.h
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import threading
import time

from .message import Message
from .network_connection import NetworkConnection

from .tcpsockets.tcp_acceptor import TcpAcceptor
from .tcpsockets.tcp_connector import TcpConnector

from .support.safe_thread import SafeThread

from .constants import (
    kKeepAlive,
    kClientHello,
    kProtoUnsup,
    kServerHello,
    kServerHelloDone,
    kClientHelloDone,
    kEntryAssign,
    NT_NET_MODE_NONE,
    NT_NET_MODE_SERVER,
    NT_NET_MODE_CLIENT,
    NT_NET_MODE_STARTING,
    NT_NET_MODE_FAILURE,
    NT_NET_MODE_TEST,
)

import logging

logger = logging.getLogger("nt")


class Dispatcher(object):
    def __init__(self, storage, conn_notifier, verbose=False):

        # logging debugging
        self.m_verbose = verbose

        self.m_storage = storage
        self.m_notifier = conn_notifier
        self.m_networkMode = NT_NET_MODE_NONE
        self.m_persist_filename = None
        self.m_server_acceptor = None
        self.m_client_connector_override = None
        self.m_client_connector = None
        self.m_connections_uid = 0

        self.m_default_proto = 0x0300  # for testing

        # Mutex for user-accessible items
        self.m_user_mutex = threading.RLock()
        self.m_connections = []

        # Circular import issue
        try:
            from .version import __version__
        except ImportError:
            __version__ = "[unknown version]"

        self.m_identity = "pynetworktables %s" % __version__

        self.m_active = False  # set to false to terminate threads
        self.m_update_rate = 0.050  # periodic dispatch rate, in s

        # Condition variable for forced dispatch wakeup (flush)
        self.m_flush_mutex = threading.Lock()
        self.m_flush_cv = threading.Condition(self.m_flush_mutex)
        self.m_last_flush = 0
        self.m_do_flush = False

        # Condition variable for client reconnect (uses user mutex)
        self.m_reconnect_cv = threading.Condition(self.m_user_mutex)
        self.m_reconnect_proto_rev = self.m_default_proto
        self.m_do_reconnect = True

    def setVerboseLogging(self, verbose):
        self.m_verbose = verbose

    def setServer(self, server_or_servers):
        """
            :param server_or_servers: a tuple of (server, port) or a list of tuples of (server, port)
        """
        self._setConnector(server_or_servers)

    def setServerTeam(self, team, port):
        servers = [
            "10.%d.%d.2" % (team / 100, team % 100),
            "roboRIO-%d-FRC.local" % team,
            "172.22.11.2",
            "roboRIO-%d-FRC.lan" % team,
            "roboRIO-%d-FRC.frc-field.local" % team,
        ]
        self.setServer([(s, port) for s in servers])

    def setServerOverride(self, server, port):
        self._setConnectorOverride((server, port))

    def clearServerOverride(self):
        self._clearConnectorOverride()

    def getNetworkMode(self):
        return self.m_networkMode

    # python-specific
    def startTestMode(self, is_server):
        with self.m_user_mutex:
            if self.m_active:
                return False
            self.m_active = True

            if is_server:
                self.m_networkMode = NT_NET_MODE_SERVER | NT_NET_MODE_TEST
            else:
                self.m_networkMode = NT_NET_MODE_CLIENT | NT_NET_MODE_TEST

        return True

    def startServer(self, persist_filename, listen_address, port):

        with self.m_user_mutex:
            if self.m_active:
                return False
            self.m_active = True

        logger.info("NetworkTables initialized in server mode")

        acceptor = TcpAcceptor(port, listen_address.strip())

        self.m_networkMode = NT_NET_MODE_SERVER | NT_NET_MODE_STARTING
        self.m_persist_filename = persist_filename
        self.m_server_acceptor = acceptor

        # Load persistent file.  Ignore errors, pass along warnings.
        if self.m_verbose:
            logger.debug("persistent filename is %s", persist_filename)

        if persist_filename:
            self.m_storage.loadPersistent(persist_filename)

        self.m_storage.setDispatcher(self, True)

        self.m_dispatch_thread = SafeThread(
            target=self._dispatchThreadMain, name="nt-dispatch-thread"
        )
        self.m_clientserver_thread = SafeThread(
            target=self._serverThreadMain, name="nt-server-thread"
        )
        return True

    def startClient(self):
        with self.m_user_mutex:
            if self.m_active:
                return False
            self.m_active = True

        logger.info("NetworkTables initialized in client mode")

        self.m_networkMode = NT_NET_MODE_CLIENT | NT_NET_MODE_STARTING
        self.m_storage.setDispatcher(self, False)

        self.m_dispatch_thread = SafeThread(
            target=self._dispatchThreadMain, name="nt-dispatch-thread"
        )
        self.m_clientserver_thread = SafeThread(
            target=self._clientThreadMain, name="nt-client-thread"
        )

        return False

    def stop(self):
        with self.m_user_mutex:
            if not self.m_active:
                return

            self.m_active = False

            # python-specific
            if self.m_networkMode & NT_NET_MODE_TEST != 0:
                return

        # wake up dispatch thread with a flush
        with self.m_flush_mutex:
            self.m_flush_cv.notify()

        # wake up client thread with a reconnect
        with self.m_user_mutex:
            self.m_client_connector = None

        self._clientReconnect()

        # wake up server thread by shutting down the socket
        if self.m_server_acceptor:
            self.m_server_acceptor.shutdown()

        # join threads, timeout
        self.m_dispatch_thread.join(1)
        self.m_clientserver_thread.join(1)

        with self.m_user_mutex:
            conns = self.m_connections
            self.m_connections = []

        # close all connections
        for conn in conns:
            conn.stop()

        # cleanup the server socket
        # -> needed because we don't have destructors
        if self.m_server_acceptor:
            self.m_server_acceptor.close()

    def setUpdateRate(self, interval):
        # don't allow update rates faster than 10 ms or slower than 1 second
        interval = float(interval)

        if interval < 0.01:
            interval = 0.01

        elif interval > 1.0:
            interval = 1.0

        self.m_update_rate = interval

    def setIdentity(self, name):
        with self.m_user_mutex:
            self.m_identity = name

    def setDefaultProtoRev(self, proto_rev):
        self.m_default_proto = proto_rev
        self.m_reconnect_proto_rev = proto_rev

    def flush(self):
        now = time.monotonic()
        with self.m_flush_mutex:
            # don't allow flushes more often than every 10 ms
            if (now - self.m_last_flush) < 0.010:
                return

            self.m_last_flush = now
            self.m_do_flush = True

            self.m_flush_cv.notify()

    def getConnections(self):
        conns = []
        if not self.m_active:
            return conns

        with self.m_user_mutex:
            for conn in self.m_connections:
                if conn.state != NetworkConnection.State.kActive:
                    continue

                conns.append(conn.info())

        return conns

    def isConnected(self):
        if self.m_active:
            with self.m_user_mutex:
                for conn in self.m_connections:
                    if conn.is_connected():
                        return True
        return False

    def isServer(self):
        return (self.m_networkMode & NT_NET_MODE_SERVER) != 0

    def addListener(self, callback, immediate_notify):
        with self.m_user_mutex:
            uid = self.m_notifier.add(callback)
            # perform immediate notifications
            if immediate_notify:
                for conn in self.m_connections:
                    if conn.is_connected():
                        self.m_notifier.notifyConnection(True, conn.info(), uid)
        return uid

    def addPolledListener(self, poller_uid, immediate_notify):
        with self.m_user_mutex:
            uid = self.m_notifier.addPolled(poller_uid)
            # perform immediate notifications
            if immediate_notify:
                for conn in self.m_connections:
                    if conn.is_connected():
                        self.m_notifier.notifyConnection(True, conn.info(), uid)
        return uid

    def _strip_connectors(self, connector):
        if isinstance(connector, tuple):
            server, port = connector
            return (server.strip(), port)
        else:
            return [(server.strip(), port) for server, port in connector]

    def _setConnector(self, connector):
        with self.m_user_mutex:
            self.m_client_connector = self._strip_connectors(connector)

    def _setConnectorOverride(self, connector):
        with self.m_user_mutex:
            self.m_client_connector_override = self._strip_connectors(connector)

    def _clearConnectorOverride(self):
        with self.m_user_mutex:
            self.m_client_connector_override = None

    def _dispatchWaitFor(self):
        return not self.m_active or self.m_do_flush

    def _dispatchThreadMain(self):
        timeout_time = time.monotonic()

        save_delta_time = 1.0
        next_save_time = timeout_time + save_delta_time

        count = 0
        is_server = self.m_networkMode & NT_NET_MODE_SERVER
        verbose = self.m_verbose

        # python micro-optimizations because this is a loop
        monotonic = time.monotonic
        kActive = NetworkConnection.State.kActive
        kDead = NetworkConnection.State.kDead

        while self.m_active:
            # handle loop taking too long
            start = monotonic()
            if start > timeout_time or timeout_time > start + self.m_update_rate:
                timeout_time = start

            # wait for periodic or when flushed
            timeout_time += self.m_update_rate
            with self.m_flush_mutex:
                self.m_flush_cv.wait_for(self._dispatchWaitFor, timeout_time - start)
                self.m_do_flush = False

            # in case we were woken up to terminate
            if not self.m_active:
                break

            # perform periodic persistent save
            if is_server and self.m_persist_filename and start > next_save_time:
                next_save_time += save_delta_time
                # handle loop taking too long
                if start > next_save_time:
                    next_save_time = start + save_delta_time

                err = self.m_storage.savePersistent(self.m_persist_filename, True)
                if err:
                    logger.warning("periodic persistent save: %s", err)

            with self.m_user_mutex:
                reconnect = False

                if verbose:
                    count += 1
                    if count > 10:
                        logger.debug(
                            "dispatch running %s connections", len(self.m_connections)
                        )
                        count = 0

                for conn in self.m_connections:
                    # post outgoing messages if connection is active
                    # only send keep-alives on client
                    state = conn.state
                    if state == kActive:
                        conn.postOutgoing(not is_server)

                    # if client, if connection died
                    if not is_server and state == kDead:
                        reconnect = True

                # reconnect if we disconnected (and a reconnect is not in progress)
                if reconnect and not self.m_do_reconnect:
                    self.m_do_reconnect = True
                    self.m_reconnect_cv.notify()

    def _queueOutgoing(self, msg, only, except_):
        with self.m_user_mutex:
            for conn in self.m_connections:
                if conn == except_:
                    continue

                if only and conn != only:
                    continue

                state = conn.state
                if (
                    state != NetworkConnection.State.kSynchronized
                    and state != NetworkConnection.State.kActive
                ):
                    continue

                conn.queueOutgoing(msg)

    def _serverThreadMain(self):
        if not self.m_server_acceptor.start():
            self.m_active = False
            self.m_networkMode = NT_NET_MODE_SERVER | NT_NET_MODE_FAILURE
            return

        self.m_networkMode = NT_NET_MODE_SERVER

        try:
            while self.m_active:
                stream = self.m_server_acceptor.accept()
                if not stream:
                    self.m_active = False
                    return

                if not self.m_active:
                    return

                logger.debug(
                    "server: client connection from %s port %s",
                    stream.getPeerIP(),
                    stream.getPeerPort(),
                )

                # add to connections list
                connection_uid = self.m_connections_uid
                self.m_connections_uid += 1
                conn = NetworkConnection(
                    connection_uid,
                    stream,
                    self.m_notifier,
                    self._serverHandshake,
                    self.m_storage.getMessageEntryType,
                    verbose=self.m_verbose,
                )

                conn.set_process_incoming(self.m_storage.processIncoming)

                with self.m_user_mutex:
                    # reuse dead connection slots
                    for i in range(len(self.m_connections)):
                        c = self.m_connections[i]
                        if c.state == NetworkConnection.State.kDead:
                            self.m_connections[i] = conn
                            break
                    else:
                        self.m_connections.append(conn)

                    conn.start()
        finally:
            self.m_networkMode = NT_NET_MODE_NONE

    def _clientThreadMain(self):
        try:
            tcp_connector = TcpConnector(1, self.m_verbose)

            while self.m_active:
                # sleep between retries
                time.sleep(0.250)
                tcp_connector.setVerbose(self.m_verbose)

                # get next server to connect to
                with self.m_user_mutex:
                    if self.m_client_connector_override:
                        server_or_servers = self.m_client_connector_override
                    else:
                        if not self.m_client_connector:
                            self.m_networkMode = (
                                NT_NET_MODE_CLIENT | NT_NET_MODE_FAILURE
                            )
                            continue
                        server_or_servers = self.m_client_connector

                # try to connect (with timeout)
                if self.m_verbose:
                    logger.debug("client trying to connect")

                stream = tcp_connector.connect(server_or_servers)
                if not stream:
                    self.m_networkMode = NT_NET_MODE_CLIENT | NT_NET_MODE_FAILURE
                    continue  # keep retrying

                logger.debug("client connected")
                self.m_networkMode = NT_NET_MODE_CLIENT

                with self.m_user_mutex:
                    connection_uid = self.m_connections_uid
                    self.m_connections_uid += 1

                    conn = NetworkConnection(
                        connection_uid,
                        stream,
                        self.m_notifier,
                        self._clientHandshake,
                        self.m_storage.getMessageEntryType,
                        verbose=self.m_verbose,
                    )

                    conn.set_process_incoming(self.m_storage.processIncoming)

                    # disconnect any current
                    # -> different from ntcore because we don't have destructors
                    for c in self.m_connections:
                        if c != conn:
                            c.stop()

                    del self.m_connections[:]
                    self.m_connections.append(conn)
                    conn.set_proto_rev(self.m_reconnect_proto_rev)
                    conn.start()

                    # reconnect the next time starting with latest protocol revision
                    self.m_reconnect_proto_rev = self.m_default_proto

                    # block until told to reconnect
                    self.m_do_reconnect = False

                    self.m_reconnect_cv.wait_for(
                        lambda: not self.m_active or self.m_do_reconnect
                    )
        finally:
            self.m_networkMode = NT_NET_MODE_NONE

    def _clientHandshake(self, conn, get_msg, send_msgs):
        # get identity
        with self.m_user_mutex:
            self_id = self.m_identity

        # send client hello
        if self.m_verbose:
            logger.debug("client: sending hello")

        send_msgs((Message.clientHello(conn.get_proto_rev(), self_id),))

        # wait for response
        msg = get_msg()
        if not msg:
            # disconnected, retry
            logger.debug("client: server disconnected before first response")
            return False

        if msg.type == kProtoUnsup:
            if msg.id == 0x0200:
                logger.debug("client: connected to NT2 server, reconnecting...")
                self._clientReconnect(0x0200)
            else:
                logger.debug("client: connected to 0x%04x server, giving up...", msg.id)

            return False

        new_server = True
        if conn.get_proto_rev() >= 0x0300:
            # should be server hello; if not, disconnect.
            if not msg.type == kServerHello:
                return False

            remote_id = msg.str
            if (msg.flags & 1) != 0:
                new_server = False

            # get the next message
            msg = get_msg()
        else:
            remote_id = "NT2 server"

        conn.set_remote_id(remote_id)

        # receive initial assignments
        incoming = []
        verbose = self.m_verbose

        while True:
            if not msg:
                # disconnected, retry
                logger.debug("client: server disconnected during initial entries")
                return False

            if msg.type == kServerHelloDone:
                break

            if msg.type == kKeepAlive:
                # shouldn't receive a keep alive, but handle gracefully
                msg = get_msg()
                continue

            if not msg.type == kEntryAssign:
                # unexpected message
                logger.debug(
                    "client: received message (%s) other than entry assignment during initial handshake",
                    msg.type,
                )
                return False

            if verbose:
                logger.debug(
                    "client %s: received assign str=%s id=%s seq_num=%s val=%s",
                    self_id,
                    msg.str,
                    msg.id,
                    msg.seq_num_uid,
                    msg.value,
                )

            incoming.append(msg)
            # get the next message
            msg = get_msg()

        # generate outgoing assignments
        outgoing = []

        self.m_storage.applyInitialAssignments(conn, incoming, new_server, outgoing)

        if conn.get_proto_rev() >= 0x0300:
            outgoing.append(Message.clientHelloDone())

        if outgoing:
            send_msgs(outgoing)

        # stream = conn.get_stream()
        # logger.info("client: CONNECTED to server %s port %s",
        #            stream.getPeerIP(), stream.getPeerPort())

        return True

    def _serverHandshake(self, conn, get_msg, send_msgs):

        verbose = self.m_verbose

        # Wait for the client to send us a hello.
        msg = get_msg()
        if not msg:
            logger.debug("server: client disconnected before sending hello")
            return False

        if not msg.type == kClientHello:
            logger.debug("server: client initial message was not client hello")
            return False

        # Check that the client requested version is not too high.
        proto_rev = msg.id
        if proto_rev > self.m_default_proto:
            logger.debug(
                "server: client requested proto > 0x%04x", self.m_default_proto
            )
            send_msgs((Message.protoUnsup(self.m_default_proto),))
            return False

        if proto_rev >= 0x0300:
            remote_id = msg.str
        else:
            remote_id = "NT2 client"

        conn.set_remote_id(remote_id)

        # Set the proto version to the client requested version
        if verbose:
            logger.debug("server: client protocol 0x%04x", proto_rev)
        conn.set_proto_rev(proto_rev)

        # Send initial set of assignments
        outgoing = []

        # Start with server hello.  TODO: initial connection flag
        if proto_rev >= 0x0300:
            with self.m_user_mutex:
                outgoing.append(Message.serverHello(0, self.m_identity))

        # Get snapshot of initial assignments
        self.m_storage.getInitialAssignments(conn, outgoing)

        # Finish with server hello done
        outgoing.append(Message.serverHelloDone())

        # Batch transmit
        if verbose:
            logger.debug("server: sending initial assignments")

        send_msgs(outgoing)

        # In proto rev 3.0 and later, handshake concludes with a client hello
        # done message, we can batch the assigns before marking the connection
        # active.  In pre-3.0, need to just immediately mark it active and hand
        # off control to the dispatcher to assign them as they arrive.
        if proto_rev >= 0x0300:
            # receive client initial assignments
            incoming = []
            while True:
                # get the next message (blocks)
                msg = get_msg()
                if not msg:
                    # disconnected, retry
                    logger.debug("server: disconnected waiting for initial entries")
                    return False

                if msg.type == kClientHelloDone:
                    break
                # shouldn't receive a keep alive, but handle gracefully
                elif msg.type == kKeepAlive:
                    continue

                if msg.type != kEntryAssign:
                    # unexpected message
                    logger.debug(
                        "server: received message (%s) other than entry assignment during initial handshake",
                        msg.type,
                    )
                    return False

                if verbose:
                    logger.debug(
                        "received assign str=%s id=%s seq_num=%s",
                        msg.str,
                        msg.id,
                        msg.seq_num_uid,
                    )

                incoming.append(msg)

            for msg in incoming:
                self.m_storage.processIncoming(msg, conn)

        # stream = conn.get_stream()
        # logger.info("server: client CONNECTED %s port %s",
        #            stream.getPeerIP(), stream.getPeerPort())

        return True

    def _clientReconnect(self, proto_rev=0x0300):
        if self.m_networkMode & NT_NET_MODE_SERVER != 0:
            return

        with self.m_user_mutex:
            self.m_reconnect_proto_rev = proto_rev
            self.m_do_reconnect = True

            self.m_reconnect_cv.notify()



from ._impl import __version__

from .entry import NetworkTableEntry
from .instance import NetworkTablesInstance
from .table import NetworkTable
from ._impl.value import Value

#: Alias of NetworkTablesInstance.getDefault(), the "default" instance
NetworkTables = NetworkTablesInstance.getDefault()

__all__ = (
    "NetworkTablesInstance",
    "NetworkTables",
    "NetworkTable",
    "NetworkTableEntry",
    "Value",
)



# notrack

__all__ = ["create_rlock", "sock_makefile"]

import socket
import threading


def create_rlock(name):
    return threading.RLock()


def sock_makefile(s, mode):
    return s.makefile(mode)


def sock_create_connection(address):
    return socket.create_connection(address)


# Call this before creating any NetworkTable objects
def enable_lock_debugging(sock_block_period=None):

    from . import _impl_debug

    _impl_debug.sock_block_period = sock_block_period

    g = globals()
    g["create_rlock"] = _impl_debug.create_tracked_rlock
    g["sock_makefile"] = _impl_debug.blocking_sock_makefile
    g["sock_create_connection"] = _impl_debug.blocking_sock_create_connection



# novalidate
"""
    Operations related to LEB128 encoding/decoding

    The algorithm is taken from Appendix C of the DWARF 3 spec. For information
    on the encodings refer to section "7.6 - Variable Length Data"

"""

import sys


def size_uleb128(value):
    count = 0
    while True:
        value >>= 7
        count += 1
        if value == 0:
            break
    return count


def encode_uleb128(value):
    out = bytearray()
    while True:
        byte = value & 0x7F
        value >>= 7
        if value != 0:
            byte = byte | 0x80
        out.append(byte)
        if value == 0:
            break
    return out


def read_uleb128(rstream):
    result = 0
    shift = 0
    while True:
        b = rstream.read(1)[0]
        result |= (b & 0x7F) << shift
        shift += 7
        if (b & 0x80) == 0:
            break
    return result



# validated: 2018-11-27 DS ac751d32247e cpp/ConnectionNotifier.cpp cpp/ConnectionNotifier.h cpp/IConnectionNotifier.h
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from collections import namedtuple

from .callback_manager import CallbackManager, CallbackThread

_ConnectionCallback = namedtuple("ConnectionCallback", ["callback", "poller_uid"])

_ConnectionNotification = namedtuple(
    "ConnectionNotification", ["connected", "conn_info"]
)


class ConnectionNotifierThread(CallbackThread):
    def __init__(self):
        CallbackThread.__init__(self, "connection-notifier")

    def matches(self, listener, data):
        return True

    def setListener(self, data, listener_uid):
        pass

    def doCallback(self, callback, data):
        callback(data)


class ConnectionNotifier(CallbackManager):

    THREAD_CLASS = ConnectionNotifierThread

    def add(self, callback):
        return self.doAdd(_ConnectionCallback(callback, None))

    def addPolled(self, poller_uid):
        return self.doAdd(_ConnectionCallback(None, poller_uid))

    def notifyConnection(self, connected, conn_info, only_listener=None):
        self.send(only_listener, _ConnectionNotification(connected, conn_info))

    def start(self):
        CallbackManager.start(self)



from typing import Any, Callable, Sequence, TypeVar, Union

from ._impl.constants import (
    NT_BOOLEAN,
    NT_DOUBLE,
    NT_STRING,
    NT_RAW,
    NT_BOOLEAN_ARRAY,
    NT_DOUBLE_ARRAY,
    NT_STRING_ARRAY,
    NT_PERSISTENT,
)

from ._impl.value import Value

__all__ = ["NetworkTableEntry"]

D = TypeVar("D")


class NetworkTableEntry:
    """
        Holds a value from NetworkTables, and changes it as new entries
        come in. Do not create this object directly, use
        :func:`.NetworkTablesInstance.getEntry` or :meth:`.NetworkTable.getEntry`
        to obtain an instance of this class.

        Using NetworkTableEntry objects to access/change NT values is more
        efficient than the getX/putX methods of :class:`.NetworkTable`.

        .. versionadded:: 2018.0.0
    """

    __slots__ = ["__api", "_local_id", "key", "_value"]

    def __init__(self, api, local_id, key):
        self.__api = api
        self._local_id = local_id

        self.key = key
        self._value = None

    def getHandle(self):
        """Gets the native handle for the entry"""
        return self._local_id

    def exists(self) -> bool:
        """Determines if the entry currently exists"""
        return self.__api.getEntryTypeById(self._local_id) != 0

    def getName(self) -> str:
        """Gets the name of the entry (the key)"""
        return self.key

    def getType(self):
        """Gets the type of the entry

        :rtype: :class:`.NetworkTablesInstance.EntryTypes`
        """
        return self.__api.getEntryTypeById(self._local_id)

    def getFlags(self) -> int:
        """Returns the flags.

        :returns: the flags (bitmask)
        """
        return self.__api.getEntryFlagsById(self._local_id)

    def getInfo(self) -> tuple:
        """Gets combined information about the entry.

        :returns: Entry information
        :rtype: tuple of (name, type, flags)
        """
        return self.__api.getEntryInfoById(self._local_id)

    @property
    def value(self):
        """Property to access the value of this entry, or None if the entry
        hasn't been initialized yet (use setXXX or forceXXX)
        """
        try:
            return self._value[1]
        except TypeError:
            return None

    # deprecated, from autoUpdateValue
    def get(self):
        try:
            return self._value[1]
        except TypeError:
            return None

    def getBoolean(self, defaultValue: D) -> Union[bool, D]:
        """Gets the entry's value as a boolean. If the entry does not exist or is of
        different type, it will return the default value.

        :param defaultValue: the value to be returned if no value is found
        :returns: the entry's value or the given default value
        """
        value = self._value
        if not value or value[0] != NT_BOOLEAN:
            return defaultValue
        return value[1]

    def getDouble(self, defaultValue: D) -> Union[float, D]:
        """Gets the entry's value as a double. If the entry does not exist or is of
        different type, it will return the default value.

        :param defaultValue: the value to be returned if no value is found
        :returns: the entry's value or the given default value
        """
        value = self._value
        if not value or value[0] != NT_DOUBLE:
            return defaultValue
        return value[1]

    getNumber = getDouble

    def getString(self, defaultValue: D) -> Union[str, D]:
        """Gets the entry's value as a string. If the entry does not exist or is of
        different type, it will return the default value.

        :param defaultValue: the value to be returned if no value is found
        :returns: the entry's value or the given default value
        """
        value = self._value
        if not value or value[0] != NT_STRING:
            return defaultValue
        return value[1]

    def getRaw(self, defaultValue: D) -> Union[bytes, D]:
        """Gets the entry's value as a raw value (byte array). If the entry does not
        exist or is of different type, it will return the default value.

        :param defaultValue: the value to be returned if no value is found
        :returns: the entry's value or the given default value
        """
        value = self._value
        if not value or value[0] != NT_RAW:
            return defaultValue
        return value[1]

    def getBooleanArray(self, defaultValue: D) -> Union[Sequence[bool], D]:
        """Gets the entry's value as a boolean array. If the entry does not
        exist or is of different type, it will return the default value.

        :param defaultValue: the value to be returned if no value is found
        :returns: the entry's value or the given default value
        """
        value = self._value
        if not value or value[0] != NT_BOOLEAN_ARRAY:
            return defaultValue
        return value[1]

    def getDoubleArray(self, defaultValue: D) -> Union[Sequence[float], D]:
        """Gets the entry's value as a double array. If the entry does not
        exist or is of different type, it will return the default value.

        :param defaultValue: the value to be returned if no value is found
        :returns: the entry's value or the given default value
        """
        value = self._value
        if not value or value[0] != NT_DOUBLE_ARRAY:
            return defaultValue
        return value[1]

    def getStringArray(self, defaultValue: D) -> Union[Sequence[str], D]:
        """Gets the entry's value as a string array. If the entry does not
        exist or is of different type, it will return the default value.

        :param defaultValue: the value to be returned if no value is found
        :returns: the entry's value or the given default value
        :rtype: list(float)
        """
        value = self._value
        if not value or value[0] != NT_STRING_ARRAY:
            return defaultValue
        return value[1]

    @classmethod
    def isValidDataType(cls, data):
        if isinstance(data, (bytes, bytearray)):
            return True
        if isinstance(data, (list, tuple)):
            if len(data) == 0:
                raise ValueError("If you use a list here, cannot be empty")
            data = data[0]

        return isinstance(data, (int, float, str, bool))

    def setDefaultValue(self, defaultValue) -> bool:
        """Sets the entry's value if it does not exist.

        :param defaultValue: the default value to set
        :returns: False if the entry exists with a different type

        .. warning:: Do not set an empty list, it will fail
        """
        value = Value.getFactory(defaultValue)(defaultValue)
        return self.__api.setDefaultEntryValueById(self._local_id, value)

    def setDefaultBoolean(self, defaultValue: bool) -> bool:
        """Sets the entry's value if it does not exist.

        :param defaultValue: the default value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeBoolean(defaultValue)
        return self.__api.setDefaultEntryValueById(self._local_id, value)

    def setDefaultDouble(self, defaultValue: float) -> bool:
        """Sets the entry's value if it does not exist.

        :param defaultValue: the default value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeDouble(defaultValue)
        return self.__api.setDefaultEntryValueById(self._local_id, value)

    setDefaultNumber = setDefaultDouble

    def setDefaultString(self, defaultValue: str) -> bool:
        """Sets the entry's value if it does not exist.

        :param defaultValue: the default value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeString(defaultValue)
        return self.__api.setDefaultEntryValueById(self._local_id, value)

    def setDefaultRaw(self, defaultValue: bytes) -> bool:
        """Sets the entry's value if it does not exist.

        :param defaultValue: the default value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeRaw(defaultValue)
        return self.__api.setDefaultEntryValueById(self._local_id, value)

    def setDefaultBooleanArray(self, defaultValue: Sequence[bool]) -> bool:
        """Sets the entry's value if it does not exist.

        :param defaultValue: the default value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeBooleanArray(defaultValue)
        return self.__api.setDefaultEntryValueById(self._local_id, value)

    def setDefaultDoubleArray(self, defaultValue: Sequence[float]) -> bool:
        """Sets the entry's value if it does not exist.

        :param defaultValue: the default value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeDoubleArray(defaultValue)
        return self.__api.setDefaultEntryValueById(self._local_id, value)

    setDefaultNumberArray = setDefaultDoubleArray

    def setDefaultStringArray(self, defaultValue: Sequence[str]) -> bool:
        """Sets the entry's value if it does not exist.

        :param defaultValue: the default value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeStringArray(defaultValue)
        return self.__api.setDefaultEntryValueById(self._local_id, value)

    def setValue(self, value) -> bool:
        """Sets the entry's value

        :param value: the value that will be assigned
        :returns: False if the table key already exists with a different type

        .. warning:: Empty lists will fail
        """
        value = Value.getFactory(value)(value)
        return self.__api.setEntryValueById(self._local_id, value)

    def setBoolean(self, value: bool) -> bool:
        """Sets the entry's value.

        :param value: the value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeBoolean(value)
        return self.__api.setEntryValueById(self._local_id, value)

    def setDouble(self, value: float) -> bool:
        """Sets the entry's value.

        :param value: the value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeDouble(value)
        return self.__api.setEntryValueById(self._local_id, value)

    setNumber = setDouble

    def setString(self, value: str) -> bool:
        """Sets the entry's value.

        :param value: the value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeString(value)
        return self.__api.setEntryValueById(self._local_id, value)

    def setRaw(self, value: bytes) -> bool:
        """Sets the entry's value.

        :param value: the value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeRaw(value)
        return self.__api.setEntryValueById(self._local_id, value)

    def setBooleanArray(self, value: Sequence[bool]) -> bool:
        """Sets the entry's value.

        :param value: the value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeBooleanArray(value)
        return self.__api.setEntryValueById(self._local_id, value)

    def setDoubleArray(self, value: Sequence[float]) -> bool:
        """Sets the entry's value.

        :param value: the value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeDoubleArray(value)
        return self.__api.setEntryValueById(self._local_id, value)

    setNumberArray = setDoubleArray

    def setStringArray(self, value: Sequence[str]) -> bool:
        """Sets the entry's value.

        :param value: the value to set
        :returns: False if the entry exists with a different type
        """
        value = Value.makeStringArray(value)
        return self.__api.setEntryValueById(self._local_id, value)

    def forceSetValue(self, value):
        """Sets the entry's value

        :param value: the value that will be assigned

        .. warning:: Empty lists will fail
        """
        value = Value.getFactory(value)(value)
        return self.__api.setEntryTypeValueById(self._local_id, value)

    def forceSetBoolean(self, value: bool):
        """Sets the entry's value.

        :param value: the value to set
        """
        value = Value.makeBoolean(value)
        return self.__api.setEntryTypeValueById(self._local_id, value)

    def forceSetDouble(self, value: float):
        """Sets the entry's value.

        :param value: the value to set
        """
        value = Value.makeDouble(value)
        return self.__api.setEntryTypeValueById(self._local_id, value)

    forceSetNumber = forceSetDouble

    def forceSetString(self, value: str):
        """Sets the entry's value.

        :param value: the value to set
        """
        value = Value.makeString(value)
        return self.__api.setEntryTypeValueById(self._local_id, value)

    def forceSetRaw(self, value: bytes):
        """Sets the entry's value.

        :param value: the value to set
        """
        value = Value.makeRaw(value)
        return self.__api.setEntryTypeValueById(self._local_id, value)

    def forceSetBooleanArray(self, value: Sequence[bool]):
        """Sets the entry's value.

        :param value: the value to set
        """
        value = Value.makeBooleanArray(value)
        return self.__api.setEntryTypeValueById(self._local_id, value)

    def forceSetDoubleArray(self, value: Sequence[float]):
        """Sets the entry's value.

        :param value: the value to set
        """
        value = Value.makeDoubleArray(value)
        return self.__api.setEntryTypeValueById(self._local_id, value)

    forceSetNumberArray = forceSetDoubleArray

    def forceSetStringArray(self, value: Sequence[str]):
        """Sets the entry's value.

        :param value: the value to set
        """
        value = Value.makeStringArray(value)
        return self.__api.setEntryTypeValueById(self._local_id, value)

    def setFlags(self, flags: int) -> None:
        """Sets flags.

        :param flags: the flags to set (bitmask)
        """
        flags = self.getFlags() | flags
        self.__api.setEntryFlagsById(self._local_id, flags)

    def clearFlags(self, flags: int) -> None:
        """Clears flags

        :param flags: the flags to clear (bitmask)
        """
        flags = self.getFlags() & ~flags
        self.__api.setEntryFlagsById(self._local_id, flags)

    def setPersistent(self) -> None:
        """Make value persistent through program restarts."""
        self.setFlags(NT_PERSISTENT)

    def clearPersistent(self) -> None:
        """Stop making value persistent through program restarts."""
        self.clearFlags(NT_PERSISTENT)

    def isPersistent(self) -> bool:
        """Returns whether the value is persistent through program restarts.

        :returns: True if the value is persistent.
        """
        return (self.getFlags() & NT_PERSISTENT) != 0

    def delete(self) -> bool:
        """Deletes the entry."""
        return self.__api.deleteEntryById(self._local_id)

    #
    # TODO: RPC entry stuff not implemented
    #

    def addListener(
        self,
        listener: Callable[["NetworkTableEntry", str, Any, int], None],
        flags: int,
        paramIsNew: bool = True,
    ):
        """Add a listener for changes to the entry

        :param listener: the listener to add
        :type listener: `callable(entry, key, value, param)`
        :param flags: bitmask specifying desired notifications
        :type flags: :class:`.NetworkTablesInstance.NotifyFlags`
        :param paramIsNew: If True, the listener fourth parameter is a boolean set to True
                           if the listener is being called because of a new value in the
                           table. Otherwise, the parameter is an integer of the raw
                           `NT_NOTIFY_*` flags

        :returns: listener handle
        """
        return self.__api.addEntryListenerByIdEx(
            self, self.key, self._local_id, listener, flags, paramIsNew
        )

    def removeListener(self, listener_id) -> None:
        """Remove a listener from receiving entry events

        :param listener: the callable that was passed to addListener
        """
        self.__api.removeEntryListener(listener_id)

    # Comparison operators et al

    def __lt__(self, other):
        raise TypeError(
            "< not allowed on NetworkTableEntry objects. Use the .value attribute instead"
        )

    def __le__(self, other):
        raise TypeError(
            "<= not allowed on NetworkTableEntry objects. Use the .value attribute instead"
        )

    def __eq__(self, other):
        raise TypeError(
            "== not allowed on NetworkTableEntry objects. Use the .value attribute instead"
        )

    def __ne__(self, other):
        raise TypeError(
            "!= not allowed on NetworkTableEntry objects. Use the .value attribute instead"
        )

    def __gt__(self, other):
        raise TypeError(
            "> not allowed on NetworkTableEntry objects. Use the .value attribute instead"
        )

    def __ge__(self, other):
        raise TypeError(
            ">= not allowed on NetworkTableEntry objects. Use the .value attribute instead"
        )

    def __bool__(self):
        raise TypeError(
            "< not allowed on NetworkTableEntry objects. Use the .value attribute instead"
        )

    def __repr__(self):
        return "<NetworkTableEntry: %s>" % (self._value.__repr__(),)



import warnings

from . import NetworkTable  # noqa

warnings.warn(
    "networktables.networktable is deprecated, import networktables.NetworkTable directly",
    DeprecationWarning,
    stacklevel=2,
)



# validated: 2017-10-01 DS e4a8bff70e77 cpp/EntryNotifier.cpp cpp/EntryNotifier.h cpp/IEntryNotifier.h
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

from collections import namedtuple

from .callback_manager import CallbackManager, CallbackThread

from .constants import (
    NT_NOTIFY_IMMEDIATE,
    NT_NOTIFY_LOCAL,
    NT_NOTIFY_UPDATE,
    NT_NOTIFY_FLAGS,
)


_EntryListenerData = namedtuple(
    "EntryListenerData",
    [
        "prefix",
        "local_id",  # we don't have entry handles like ntcore has
        "flags",
        "callback",
        "poller_uid",
    ],
)

#
_EntryNotification = namedtuple(
    "EntryNotification", ["name", "value", "flags", "local_id"]
)

_assign_both = NT_NOTIFY_UPDATE | NT_NOTIFY_FLAGS
_immediate_local = NT_NOTIFY_IMMEDIATE | NT_NOTIFY_LOCAL


class EntryNotifierThread(CallbackThread):
    def __init__(self):
        CallbackThread.__init__(self, "entry-notifier")

    def matches(self, listener, data):
        if not data.value:
            return False

        # must match local id or prefix
        # -> python-specific: match this first, since it's the most likely thing
        #    to not match
        if listener.local_id is not None:
            if listener.local_id != data.local_id:
                return False
        else:
            if not data.name.startswith(listener.prefix):
                return False

        # Flags must be within requested flag set for this listener.
        # Because assign messages can result in both a value and flags update,
        # we handle that case specially.
        listen_flags = listener.flags & ~_immediate_local
        flags = data.flags & ~_immediate_local

        if (flags & _assign_both) == _assign_both:
            if (listen_flags & _assign_both) == 0:
                return False
            listen_flags &= ~_assign_both
            flags &= ~_assign_both

        if (flags & ~listen_flags) != 0:
            return False

        return True

    def setListener(self, data, listener_uid):
        pass

    def doCallback(self, callback, data):
        callback(data)


class EntryNotifier(CallbackManager):

    THREAD_CLASS = EntryNotifierThread

    def __init__(self, verbose):
        CallbackManager.__init__(self, verbose)

        self.m_local_notifiers = False

    def add(self, callback, prefix, flags):
        if (flags & NT_NOTIFY_LOCAL) != 0:
            self.m_local_notifiers = True
        return self.doAdd(_EntryListenerData(prefix, None, flags, callback, None))

    def addById(self, callback, local_id, flags):
        if (flags & NT_NOTIFY_LOCAL) != 0:
            self.m_local_notifiers = True
        return self.doAdd(_EntryListenerData(None, local_id, flags, callback, None))

    def addPolled(self, poller_uid, prefix, flags):
        if (flags & NT_NOTIFY_LOCAL) != 0:
            self.m_local_notifiers = True
        return self.doAdd(_EntryListenerData(prefix, None, flags, None, poller_uid))

    def addPolledById(self, poller_uid, local_id, flags):
        if (flags & NT_NOTIFY_LOCAL) != 0:
            self.m_local_notifiers = True
        return self.doAdd(_EntryListenerData(None, local_id, flags, None, poller_uid))

    def notifyEntry(self, local_id, name, value, flags, only_listener=None):

        # optimization: don't generate needless local queue entries if we have
        # no local listeners (as this is a common case on the server side)
        if not self.m_local_notifiers and (flags & NT_NOTIFY_LOCAL) != 0:
            return

        self.send(only_listener, _EntryNotification(name, value, flags, local_id))

    def start(self):
        CallbackManager.start(self)



# novalidate

import socket
import threading

from .tcp_stream import TCPStream

import logging

logger = logging.getLogger("nt.net")


class TcpConnector(object):
    def __init__(self, timeout, verbose):
        self.cond = threading.Condition()
        self.threads = {}
        self.active = False
        self.result = None
        self.timeout = timeout
        self.verbose = verbose

    def setVerbose(self, verbose):
        self.verbose = verbose

    def connect(self, server_or_servers):
        if isinstance(server_or_servers, tuple):
            server, port = server_or_servers
            return self._connect(server, port)

        # parallel connect
        # -> only connect to servers that aren't currently being connected to
        with self.cond:
            self.active = True
            for item in server_or_servers:
                if item not in self.threads:
                    th = threading.Thread(
                        target=self._thread, args=item, name="TcpConnector"
                    )
                    th.daemon = True
                    th.start()
                    self.threads[item] = th

            self.cond.wait(self.timeout)
            self.active = False

            result = self.result
            self.result = None
            return result

    def _thread(self, server, port):
        stream = self._connect(server, port)
        with self.cond:
            self.threads.pop((server, port), None)
            if self.active and self.result is None:
                self.result = stream
                self.cond.notify()

    def _connect(self, server, port):
        try:
            if self.verbose:
                logger.debug("Trying connection to %s:%s", server, port)

            if self.timeout is None:
                sd = socket.create_connection((server, port))
            else:
                sd = socket.create_connection((server, port), timeout=self.timeout)
                sd.settimeout(None)

            return TCPStream(sd, server, port, "client")
        except IOError:
            if self.verbose:
                logger.debug("Connection to %s:%s failed", server, port)
            return



# validated: 2018-11-27 DS 18c8cce6a78d cpp/DsClient.cpp cpp/DsClient.h
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import json
import threading

from .support.safe_thread import SafeThread
from .tcpsockets.tcp_connector import TcpConnector

import logging

logger = logging.getLogger("nt")


class DsClient(object):
    def __init__(self, dispatcher, verbose=False):
        self.m_dispatcher = dispatcher
        self.verbose = verbose

        self.m_active = False
        self.m_owner = None  # type: SafeThread

        self.m_mutex = threading.Lock()
        self.m_cond = threading.Condition(self.m_mutex)

        self.m_port = None  # type: int
        self.m_stream = None

    def start(self, port):
        with self.m_mutex:
            self.m_port = port
            if not self.m_active:
                self.m_active = True
                self.m_owner = SafeThread(target=self._thread_main, name="nt-dsclient")

    def stop(self):
        with self.m_mutex:
            # Close the stream so the read (if any) terminates.
            self.m_active = False
            if self.m_stream:
                self.m_stream.close()
            self.m_cond.notify()

    def _thread_main(self):
        oldip = 0
        connector = TcpConnector(verbose=False, timeout=1)

        while self.m_active:
            # wait for periodic reconnect or termination
            with self.m_mutex:
                self.m_cond.wait_for(lambda: not self.m_active, timeout=0.5)
                port = self.m_port

            if not self.m_active:
                break

            self.m_stream = connector.connect(("127.0.0.1", 1742))
            if not self.m_active:
                break
            if not self.m_stream:
                continue

            while self.m_active and self.m_stream:
                json_blob = self.m_stream.readline()
                if not json_blob:
                    # We've reached EOF.
                    with self.m_mutex:
                        self.m_stream.close()
                        self.m_stream = None

                if not self.m_active:
                    break

                try:
                    obj = json.loads(json_blob.decode())
                except (json.JSONDecodeError, UnicodeDecodeError):
                    continue
                try:
                    ip = int(obj["robotIP"])
                except (KeyError, ValueError):
                    continue

                # If zero, clear the server override
                if ip == 0:
                    self.m_dispatcher.clearServerOverride()
                    oldip = 0
                    continue

                # If unchanged, don't reconnect
                if ip == oldip:
                    continue
                oldip = ip

                # Convert number into dotted quad
                ip_str = "%d.%d.%d.%d" % (
                    (ip >> 24) & 0xFF,
                    (ip >> 16) & 0xFF,
                    (ip >> 8) & 0xFF,
                    ip & 0xFF,
                )
                if self.verbose:
                    logger.info("client: DS overriding server IP to %s", ip_str)
                self.m_dispatcher.setServerOverride(ip_str, port)

            # We disconnected from the DS, clear the server override
            self.m_dispatcher.clearServerOverride()
            oldip = 0

        # Python note: we don't call Dispatcher.clearServerOverride() again.
        # Either it was already called, or we were never active.



# validated: 2017-09-28 DS cedbafeb286d cpp/NetworkConnection.cpp cpp/NetworkConnection.h cpp/INetworkConnection.h
# ----------------------------------------------------------------------------
# Copyright (c) FIRST 2017. All Rights Reserved.
# Open Source Software - may be modified and shared by FRC teams. The code
# must be accompanied by the FIRST BSD license file in the root directory of
# the project.
# ----------------------------------------------------------------------------

import threading
from time import monotonic

try:
    # Python 3.7 only, should be more efficient
    from queue import SimpleQueue as Queue, Empty
except ImportError:
    from queue import Queue, Empty


from .constants import (
    kEntryAssign,
    kEntryUpdate,
    kFlagsUpdate,
    kEntryDelete,
    kClearEntries,
    msgtype_str,
)

from .message import Message
from .structs import ConnectionInfo
from .wire import WireCodec

from .support.lists import Pair
from .support.safe_thread import SafeThread

from .tcpsockets.tcp_stream import StreamEOF

import logging

logger = logging.getLogger("nt")

_empty_pair = Pair(0, 0)

_state_map = {
    0: "created",
    1: "init",
    2: "handshake",
    3: "synchronized",
    4: "active",
    5: "dead",
}


class NetworkConnection(object):
    class State(object):
        kCreated = 0
        kInit = 1
        kHandshake = 2
        kSynchronized = 3
        kActive = 4
        kDead = 5

    def __init__(self, uid, stream, notifier, handshake, get_entry_type, verbose=False):

        # logging debugging
        self.m_verbose = verbose

        self.m_uid = uid
        self.m_stream = stream
        self.m_notifier = notifier
        self.m_handshake = handshake
        self.m_get_entry_type = get_entry_type

        self.m_active = False
        self.m_proto_rev = 0x0300
        self.state = self.State.kCreated
        self.m_state_mutex = threading.Lock()
        self.m_last_update = 0

        self.m_outgoing = Queue()

        self.m_process_incoming = None
        self.m_read_thread = None
        self.m_write_thread = None

        self.m_remote_id_mutex = threading.Lock()
        self.m_remote_id = None
        self.m_last_post = 0

        self.m_pending_mutex = threading.Lock()
        self.m_pending_outgoing = []
        self.m_pending_update = {}

        # Condition variables for shutdown
        self.m_shutdown_mutex = threading.Lock()
        # Not needed in python
        # self.m_read_shutdown_cv = threading.Condition()
        # self.m_write_shutdown_cv = threading.Condition()
        self.m_read_shutdown = False
        self.m_write_shutdown = False

        # turn off Nagle algorithm; we bundle packets for transmission
        try:
            self.m_stream.setNoDelay()
        except IOError as e:
            logger.warning("Setting TCP_NODELAY: %s", e)

    def start(self):
        if self.m_active:
            return

        self.m_active = True
        self.set_state(self.State.kInit)

        # clear queue
        try:
            while True:
                self.m_outgoing.get_nowait()
        except Empty:
            pass

        # reset shutdown flags
        with self.m_shutdown_mutex:
            self.m_read_shutdown = False
            self.m_write_shutdown = False

        # start threads
        self.m_write_thread = SafeThread(
            target=self._writeThreadMain, name="nt-net-write"
        )
        self.m_read_thread = SafeThread(target=self._readThreadMain, name="nt-net-read")

    def __repr__(self):
        try:
            return "<NetworkConnection 0x%x %s>" % (id(self), self.info())
        except Exception:
            return "<NetworkConnection 0x%x ???>" % id(self)

    def stop(self):
        logger.debug("NetworkConnection stopping (%s)", self)

        if not self.m_active:
            return

        self.set_state(self.State.kDead)
        self.m_active = False
        # closing the stream so the read thread terminates
        self.m_stream.close()

        # send an empty outgoing message set so the write thread terminates
        self.m_outgoing.put([])

        # wait for threads to terminate, timeout
        self.m_write_thread.join(1)
        if self.m_write_thread.is_alive():
            logger.warning("%s did not die", self.m_write_thread.name)

        self.m_read_thread.join(1)
        if self.m_read_thread.is_alive():
            logger.warning("%s did not die", self.m_write_thread.name)

        # clear queue
        try:
            while True:
                self.m_outgoing.get_nowait()
        except Empty:
            pass

    def get_proto_rev(self):
        return self.m_proto_rev

    def get_stream(self):
        return self.m_stream

    def info(self):
        return ConnectionInfo(
            self.remote_id(),
            self.m_stream.getPeerIP(),
            self.m_stream.getPeerPort(),
            self.m_last_update,
            self.m_proto_rev,
        )

    def is_connected(self):
        return self.state == self.State.kActive

    def last_update(self):
        return self.m_last_update

    def set_process_incoming(self, func):
        self.m_process_incoming = func

    def set_proto_rev(self, proto_rev):
        self.m_proto_rev = proto_rev

    def set_state(self, state):
        with self.m_state_mutex:
            State = self.State

            # Don't update state any more once we've died
            if self.state == State.kDead:
                return

            # One-shot notify state changes
            if self.state != State.kActive and state == State.kActive:
                info = self.info()
                self.m_notifier.notifyConnection(True, info)
                logger.info(
                    "CONNECTED %s port %s (%s)",
                    info.remote_ip,
                    info.remote_port,
                    info.remote_id,
                )
            elif self.state != State.kDead and state == State.kDead:
                info = self.info()
                self.m_notifier.notifyConnection(False, info)
                logger.info(
                    "DISCONNECTED %s port %s (%s)",
                    info.remote_ip,
                    info.remote_port,
                    info.remote_id,
                )

            if self.m_verbose:
                logger.debug(
                    "%s: %s -> %s", self, _state_map[self.state], _state_map[state]
                )

            self.state = state

    # python optimization: don't use getter here
    # def state(self):
    #     return self.m_state

    def remote_id(self):
        with self.m_remote_id_mutex:
            return self.m_remote_id

    def set_remote_id(self, remote_id):
        with self.m_remote_id_mutex:
            self.m_remote_id = remote_id

    def uid(self):
        return self.m_uid

    def _sendMessages(self, msgs):
        self.m_outgoing.put(msgs)

    def _readThreadMain(self):
        decoder = WireCodec(self.m_proto_rev)

        verbose = self.m_verbose

        def _getMessage():
            decoder.set_proto_rev(self.m_proto_rev)
            try:
                return Message.read(self.m_stream, decoder, self.m_get_entry_type)
            except IOError as e:
                logger.warning("read error in handshake: %s", e)

                # terminate connection on bad message
                self.m_stream.close()

                return None

        self.set_state(self.State.kHandshake)

        try:
            handshake_success = self.m_handshake(self, _getMessage, self._sendMessages)
        except Exception:
            logger.exception("Unhandled exception during handshake")
            handshake_success = False

        if not handshake_success:
            self.set_state(self.State.kDead)
            self.m_active = False
        else:
            self.set_state(self.State.kActive)

            try:
                while self.m_active:
                    if not self.m_stream:
                        break

                    decoder.set_proto_rev(self.m_proto_rev)

                    try:
                        msg = Message.read(
                            self.m_stream, decoder, self.m_get_entry_type
                        )
                    except Exception as e:
                        if not isinstance(e, StreamEOF):
                            if verbose:
                                logger.exception("read error")
                            else:
                                logger.warning("read error: %s", e)

                        # terminate connection on bad message
                        self.m_stream.close()

                        break

                    if verbose:
                        logger.debug(
                            "%s received type=%s with str=%s id=%s seq_num=%s value=%s",
                            self.m_stream.sock_type,
                            msgtype_str(msg.type),
                            msg.str,
                            msg.id,
                            msg.seq_num_uid,
                            msg.value,
                        )

                    self.m_last_update = monotonic()
                    self.m_process_incoming(msg, self)
            except IOError as e:
                # connection died probably
                logger.debug("IOError in read thread: %s", e)
            except Exception:
                logger.warning("Unhandled exception in read thread", exc_info=True)

            self.set_state(self.State.kDead)
            self.m_active = False

        # also kill write thread
        self.m_outgoing.put([])

        with self.m_shutdown_mutex:
            self.m_read_shutdown = True

    def _writeThreadMain(self):
        encoder = WireCodec(self.m_proto_rev)

        verbose = self.m_verbose
        out = []

        try:
            while self.m_active:
                msgs = self.m_outgoing.get()

                if verbose:
                    logger.debug("write thread woke up")
                    if msgs:
                        logger.debug(
                            "%s sending %s messages", self.m_stream.sock_type, len(msgs)
                        )

                if not msgs:
                    continue

                encoder.set_proto_rev(self.m_proto_rev)

                # python-optimization: checking verbose causes extra overhead
                if verbose:
                    for msg in msgs:
                        if msg:
                            logger.debug(
                                "%s sending type=%s with str=%s id=%s seq_num=%s value=%s",
                                self.m_stream.sock_type,
                                msgtype_str(msg.type),
                                msg.str,
                                msg.id,
                                msg.seq_num_uid,
                                msg.value,
                            )
                            Message.write(msg, out, encoder)
                else:
                    for msg in msgs:
                        if msg:
                            Message.write(msg, out, encoder)

                if not self.m_stream:
                    break

                if not out:
                    continue

                self.m_stream.send(b"".join(out))

                del out[:]

                # if verbose:
                #    logger.debug('send %s bytes', encoder.size())
        except IOError as e:
            # connection died probably
            if not isinstance(e, StreamEOF):
                logger.debug("IOError in write thread: %s", e)
        except Exception:
            logger.warning("Unhandled exception in write thread", exc_info=True)

        self.set_state(self.State.kDead)
        self.m_active = False
        self.m_stream.close()  # also kill read thread

        with self.m_shutdown_mutex:
            self.m_write_shutdown = True

    def queueOutgoing(self, msg):
        with self.m_pending_mutex:

            # Merge with previous.  One case we don't combine: delete/assign loop.
            msgtype = msg.type
            if msgtype in [kEntryAssign, kEntryUpdate]:

                # don't do this for unassigned id's
                msg_id = msg.id
                if msg_id == 0xFFFF:
                    self.m_pending_outgoing.append(msg)
                    return

                mpend = self.m_pending_update.get(msg_id)
                if mpend is not None and mpend.first != 0:
                    # overwrite the previous one for this id
                    oldidx = mpend.first - 1
                    oldmsg = self.m_pending_outgoing[oldidx]
                    if (
                        oldmsg
                        and oldmsg.type == kEntryAssign
                        and msgtype == kEntryUpdate
                    ):
                        # need to update assignment with seq_num and value
                        oldmsg = Message.entryAssign(
                            oldmsg.str, msg_id, msg.seq_num_uid, msg.value, oldmsg.flags
                        )

                    else:
                        oldmsg = msg  # easy update

                    self.m_pending_outgoing[oldidx] = oldmsg

                else:
                    # new, remember it
                    pos = len(self.m_pending_outgoing)
                    self.m_pending_outgoing.append(msg)
                    self.m_pending_update[msg_id] = Pair(pos + 1, 0)

            elif msgtype == kEntryDelete:
                # don't do this for unassigned id's
                msg_id = msg.id
                if msg_id == 0xFFFF:
                    self.m_pending_outgoing.append(msg)
                    return

                # clear previous updates
                mpend = self.m_pending_update.get(msg_id)
                if mpend is not None:
                    if mpend.first != 0:
                        self.m_pending_outgoing[mpend.first - 1] = None

                    if mpend.second != 0:
                        self.m_pending_outgoing[mpend.second - 1] = None

                    self.m_pending_update[msg_id] = _empty_pair

                # add deletion
                self.m_pending_outgoing.append(msg)

            elif msgtype == kFlagsUpdate:
                # don't do this for unassigned id's
                msg_id = msg.id
                if id == 0xFFFF:
                    self.m_pending_outgoing.append(msg)
                    return

                mpend = self.m_pending_update.get(msg_id)
                if mpend is not None and mpend.second != 0:
                    # overwrite the previous one for this id
                    self.m_pending_outgoing[mpend.second - 1] = msg

                else:
                    # new, remember it
                    pos = len(self.m_pending_outgoing)
                    self.m_pending_outgoing.append(msg)
                    self.m_pending_update[msg_id] = Pair(0, pos + 1)

            elif msgtype == kClearEntries:
                # knock out all previous assigns/updates!
                for i, m in enumerate(self.m_pending_outgoing):
                    if not m:
                        continue

                    t = m.type
                    if t in [
                        kEntryAssign,
                        kEntryUpdate,
                        kFlagsUpdate,
                        kEntryDelete,
                        kClearEntries,
                    ]:
                        self.m_pending_outgoing[i] = None

                self.m_pending_update.clear()
                self.m_pending_outgoing.append(msg)

            else:
                self.m_pending_outgoing.append(msg)

    def postOutgoing(self, keep_alive):
        with self.m_pending_mutex:
            # optimization: don't call monotonic unless needed
            # now = monotonic()
            if not self.m_pending_outgoing:
                if not keep_alive:
                    return

                # send keep-alives once a second (if no other messages have been sent)
                now = monotonic()
                if (now - self.m_last_post) < 1.0:
                    return

                self.m_outgoing.put((Message.keepAlive(),))

            else:
                now = monotonic()
                self.m_outgoing.put(self.m_pending_outgoing)

                self.m_pending_outgoing = []
                self.m_pending_update.clear()

            self.m_last_post = now



# validated: 2018-11-27 DS 175c6c1f0130 cpp/Value.cpp include/networktables/NetworkTableValue.h
"""
    Internal storage for ntcore values

    Uses namedtuple for efficiency, and because Value objects are supposed
    to be immutable. Will have to measure that and see if there's a performance
    penalty for this in python.

    Original ntcore stores the last change time, but it doesn't seem to
    be used anywhere, so we don't store that to make equality comparison
    more efficient.
"""

from collections import namedtuple
from .constants import (
    NT_BOOLEAN,
    NT_DOUBLE,
    NT_STRING,
    NT_RAW,
    NT_BOOLEAN_ARRAY,
    NT_DOUBLE_ARRAY,
    NT_STRING_ARRAY,
    NT_RPC,
)


class Value(namedtuple("Value", ["type", "value"])):
    __slots__ = ()

    @classmethod
    def makeBoolean(cls, value):
        if value:
            return cls._TRUE_VALUE
        else:
            return cls._FALSE_VALUE

    @classmethod
    def makeDouble(cls, value):
        return cls(NT_DOUBLE, float(value))

    @classmethod
    def makeString(cls, value):
        return cls(NT_STRING, str(value))

    @classmethod
    def makeRaw(cls, value):
        return cls(NT_RAW, bytes(value))

    # TODO: array stuff a good idea?

    @classmethod
    def makeBooleanArray(cls, value):
        return cls(NT_BOOLEAN_ARRAY, tuple(bool(v) for v in value))

    @classmethod
    def makeDoubleArray(cls, value):
        return cls(NT_DOUBLE_ARRAY, tuple(float(v) for v in value))

    @classmethod
    def makeStringArray(cls, value):
        return cls(NT_STRING_ARRAY, tuple(str(v) for v in value))

    @classmethod
    def makeRpc(cls, value):
        return cls(NT_RPC, str(value))

    @classmethod
    def getFactory(cls, value):
        if isinstance(value, bool):
            return cls.makeBoolean
        elif isinstance(value, (int, float)):
            return cls.makeDouble
        elif isinstance(value, str):
            return cls.makeString
        elif isinstance(value, (bytes, bytearray)):
            return cls.makeRaw

        # Do best effort for arrays, but can't catch all cases
        # .. if you run into an error here, use a less generic type
        elif isinstance(value, (list, tuple)):
            if not value:
                raise TypeError("If you use a list here, cannot be empty")

            first = value[0]
            if isinstance(first, bool):
                return cls.makeBooleanArray
            elif isinstance(first, (int, float)):
                return cls.makeDoubleArray
            elif isinstance(first, str):
                return cls.makeStringArray
            else:
                raise ValueError("Can only use lists of bool/int/float/strings")

        elif value is None:
            raise ValueError("Cannot put None into NetworkTable")
        else:
            raise ValueError(
                "Can only put bool/int/float/str/bytes or lists/tuples of them"
            )

    @classmethod
    def getFactoryByType(cls, type_id):
        return cls._make_map[type_id]


# optimization
Value._TRUE_VALUE = Value(NT_BOOLEAN, True)
Value._FALSE_VALUE = Value(NT_BOOLEAN, False)

Value._make_map = {
    NT_BOOLEAN: Value.makeBoolean,
    NT_DOUBLE: Value.makeDouble,
    NT_STRING: Value.makeString,
    NT_RAW: Value.makeRaw,
    NT_BOOLEAN_ARRAY: Value.makeBooleanArray,
    NT_DOUBLE_ARRAY: Value.makeDoubleArray,
    NT_STRING_ARRAY: Value.makeStringArray,
    NT_RPC: Value.makeRpc,
}
